/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/
#include "sock.h"

FDOPS sockops = { do_poll, do_close };

int
sys_socket (HVM hVM, PCRS pRegs)
{
	PSOCKET pSock;
	int res;

	if ((pRegs->Client_EBX != 17) &&
		(pRegs->Client_EBX != 6))
		return -SOCKERR_INVALID_PARAM;

	pSock = NewSocket();
	if (!pSock)
		return -SOCKERR_NO_MEMORY;

	pSock->s_recvbuf = NewCBuf (32 * 1024);
	if (!pSock->s_recvbuf) {
		FreeSocket (pSock);
		return -SOCKERR_NO_MEMORY;
	}

	pSock->s_flags = pRegs->Client_EBX == 17 ? SOCKFL_UDP : SOCKFL_TCP;
	/* XXX hack -- set UDP sockets to nonblocking mode by default */
	if (pSock->s_flags & SOCKFL_UDP)
		pSock->s_flags |= SOCKFL_NONBLOCK;
	pSock->s_sendmax = 5000;
	pSock->s_sendlowwater = 1024;

	res = AllocateFD (hVM, pRegs->Client_EDI, &sockops, (PVOID) pSock);
	if (res) {
		FreeSocket (pSock);
		return res;
	}

	return 0;
}

#define GET_NONBLOCK 0
#define SET_NONBLOCK 1

int
sys_getsetmode (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	switch (pRegs->Client_ECX) {
	case GET_NONBLOCK:
		pRegs->Client_EDX = pSock->s_flags & SOCKFL_NONBLOCK ? 1 : 0;
		break;

	case SET_NONBLOCK:
		if (pRegs->Client_EDX)
			pSock->s_flags |= SOCKFL_NONBLOCK;
		else
			pSock->s_flags &= ~SOCKFL_NONBLOCK;
		break;

	default:
		return -SOCKERR_INVALID_PARAM;
	}
	return 0;
}


void
DisconnectCallback (PVOID context, TDI_STATUS finalstatus, unsigned long bytes)
{
	PSOCKET pSock = (PSOCKET) context;

	if (!pSock || (pSock->s_signature != SOCKSIG))
		return;

	FreeSocket(pSock);
}

int
sys_close (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	TDI_STATUS status = TDI_SUCCESS;
#if 0
	/* XXX- this should work to give us graceful disconnects... why not? */
	TDI_REQUEST r;

	if (pSock->s_flags & SOCKFL_CONNECTED) {
		r.Handle.ConnectionContext = pSock->s_conhdl;
		r.RequestNotifyObject = &DisconnectCallback;
		r.RequestContext = (PVOID) pSock;
		status = pDT->TdiDisconnectEntry (&r, NULL, 0, NULL, NULL);
	}
#endif

	DeallocateFD(hVM, pRegs->Client_EDI);

	if (status != TDI_PENDING)
		FreeSocket(pSock);

	return (status == TDI_SUCCESS || status == TDI_PENDING) ? 0 : -status;
}

void
do_close (PVOID p)
{
	PSOCKET pSock = (PSOCKET) p;

	if (!pSock || (pSock->s_signature != SOCKSIG))
		return;

	FreeSocket(pSock);
}

DWORD
do_poll (PVOID context, PDWORD post)
{
	PSOCKET pSock = (PSOCKET) context;
	DWORD res = 0;

	pSock->s_post = post;
	if (pSock->s_flags & SOCKFL_UDP) {
		if (pSock->s_recvpackets)
			res |= POLL_READ;
		res |= POLL_WRITE;
	} else {
		if (pSock->s_flags & SOCKFL_LISTENING) {
			PSOCKET p;
			for (p = pSock->s_next; p; p = p->s_next) {
				if (p->s_flags & SOCKFL_CONNECTED) {
					res |= POLL_READ;
					break;
				}
			}
		} else if (pSock->s_flags & SOCKFL_CONNECTED) {
			if (pSock->s_recvbuf && !EmptyCBuf (pSock->s_recvbuf))
				res |= POLL_READ;
			if (pSock->s_flags & SOCKFL_PEERCLOSED)
				res |= POLL_READ;
			if (pSock->s_flags & SOCKFL_PEERABORTED)
				res |= POLL_EXCEPTION;
			if (pSock->s_sendbytes <= pSock->s_sendlowwater)
				res |= POLL_WRITE;
		}
	}
	return res;
}

char reuseopt[] = "\001\0";

int do_bind (HVM hVM, PCRS pRegs, PSOCKET pSock, PTRANSPORT_ADDRESS pTA)
{
	TDI_REQUEST r;
	PCALLBACKMEMO pCBM = NewCallback();
	PADDRESSOBJECT pAddr = NewAddressObject();
	int res = 0;
	TDI_STATUS err;

	if (pSock->s_flags & SOCKFL_BOUND) {
		res = -SOCKERR_ALREADY_BOUND;
		goto error;
	}

	if (!pTA || !pCBM || !pAddr) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiOpenAddressEntry (&r, pTA, (pSock->s_flags & SOCKFL_TCP ? 6 : 17), 
			(PVOID) reuseopt);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	res = err;
	if (err == TDI_SUCCESS) {
		pSock->s_flags |= SOCKFL_BOUND;
		pSock->s_addrobj = pAddr;
		pAddr->addr_hdl = r.Handle.AddressHandle;
		res = pDT->TdiSetEventEntry ((PVOID) pAddr->addr_hdl, 
			TDI_EVENT_RECEIVE_DATAGRAM, RcvDatagram, (PVOID) pSock);
		res = pDT->TdiSetEventEntry ((PVOID) pAddr->addr_hdl,
			TDI_EVENT_RECEIVE, Receive, NULL);
#if 0
		res = pDT->TdiSetEventEntry ((PVOID) pAddr->addr_hdl,
			TDI_EVENT_CONNECT, ConnectEvent, (PVOID) pSock);
#endif
		res = pDT->TdiSetEventEntry ((PVOID) pAddr->addr_hdl,
			TDI_EVENT_DISCONNECT, DisconnectEvent, (PVOID) pSock);
	}

	if (pCBM) FreeCallback(pCBM);
	return res;

error:
	if (pCBM) FreeCallback(pCBM);
	if (pAddr) ReleaseAddress(pAddr);
	return res;
}


int
sys_bind (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	PTRANSPORT_ADDRESS pTA = MakeTransportAddress(pRegs->Client_EBX, (WORD) pRegs->Client_EDX);
	int res;

	if (!pTA)
		return -SOCKERR_NO_MEMORY;

	res = do_bind (hVM, pRegs, pSock, pTA);
	FreeTransportAddress(pTA);
	return res;
}

extern void TimeOutEventHandler();

int
sys_select (HVM hVM, PCRS pRegs)
{
	PFDTABLE *p = (PFDTABLE *)((DWORD) hVM + cbmyVMarea);
	PFDTABLE pfd = *p;
	DWORD rfdin, wfdin, efdin;
	DWORD rfdout, wfdout, efdout;
	DWORD allfd, bit, pollres;
	PDWORD ptime;
	DWORD timeo, starttime, elapsedtime;
	int i, numpolls, numloops;

	if (!pfd)
		return -SOCKERR_INVALID_SOCKET;

	rfdin = pRegs->Client_EBX;
	wfdin = pRegs->Client_ECX;
	efdin = pRegs->Client_EDX;
	rfdout = wfdout = efdout = 0;
	timeo = pRegs->Client_EDI;
	pfd->fd_tei.post = &pfd->fd_post;
	pfd->fd_tei.timeeventhdl = 0;
	if (timeo && (timeo != 0xffffffff)) {
		ptime = Get_System_Time_Address();
		starttime = *ptime;
	}

	allfd = rfdin | wfdin | efdin;
	if (allfd & (~pfd->fd_inuse))
		return -SOCKERR_INVALID_SOCKET;

	if (!allfd)
		goto done;

	numloops = 0;
	while (1) {
		numpolls = 0;
		numloops++;
		pfd->fd_post = 0;
		for (i = 0; i < 32; i++) {
			bit = (1<<i);
			if (allfd & bit) {
				if (pfd->fd_fd[i].fd_ops &&
					pfd->fd_fd[i].fd_ops->fdop_poll) {
					pollres = (*pfd->fd_fd[i].fd_ops->fdop_poll) (pfd->fd_fd[i].fd_context, 
									&pfd->fd_post);
					numpolls++;
				} else
					pollres = 0;
				if ((pollres & POLL_READ) && (rfdin & bit))
					rfdout |= bit;
				if ((pollres & POLL_WRITE) && (wfdin & bit))
					wfdout |= bit;
				if ((pollres & POLL_EXCEPTION) && (efdin & bit))
					efdout |= bit;
			}
		}
		if (rfdout || wfdout || efdout)
			break;
		if (timeo == 0)
			break;
		if (timeo && (timeo != 0xffffffff)) {
			elapsedtime = *ptime - starttime;
			if (elapsedtime >= timeo)
				break;
			Cancel_Time_Out(pfd->fd_tei.timeeventhdl);
			pfd->fd_tei.timeeventhdl = Set_Global_Time_Out(TimeOutEventHandler, 
				timeo - elapsedtime, (DWORD) &pfd->fd_tei);
		}
		if (numpolls == 0)
			return -SOCKERR_INFINITE_WAIT;
		_asm pushf
		_asm cli
		if (!pfd->fd_post) 
			_BlockOnID ((DWORD) &pfd->fd_post, 
				BLOCK_THREAD_IDLE | BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
		_asm popf
	}

	for (i = 0; i < 32; i++) {
		bit = 1<<i;
		if (allfd & bit) {
			if (pfd->fd_fd[i].fd_ops &&
				pfd->fd_fd[i].fd_ops->fdop_poll)
				pollres = (*pfd->fd_fd[i].fd_ops->fdop_poll) (pfd->fd_fd[i].fd_context, 
								NULL);
		}
	}

done:
	pRegs->Client_EBX = rfdout;
	pRegs->Client_ECX = wfdout;
	pRegs->Client_EDX = efdout;
	if (timeo && (timeo != 0xffffffff)) {
		elapsedtime = *ptime - starttime;
		pRegs->Client_EDI = timeo > elapsedtime ? timeo-elapsedtime : 0;
	}
	Cancel_Time_Out (pfd->fd_tei.timeeventhdl);
	pfd->fd_tei.timeeventhdl = 0;
	return 0;
}

void
wait (PSOCKET pSock, DWORD pollmask)
{
	DWORD pollres;
	DWORD waitvar;

	while (1) {
		waitvar = 0;
		pollres = do_poll ((PVOID) pSock, &waitvar);
		if (pollres & pollmask)
			break;
		_asm pushf
		_asm cli
		if (!waitvar) 
			_BlockOnID ((DWORD) &waitvar, 
				BLOCK_THREAD_IDLE | BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);
		_asm popf
	}

	do_poll((PVOID) pSock, NULL);
}

int
sys_getname (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	PNDIS_BUFFER pNB = _HeapAllocate(sizeof(NDIS_BUFFER) + sizeof(TDI_ADDRESS_INFO) + 200,
							HEAPZEROINIT);
	PCALLBACKMEMO pCBM = NewCallback();
	PTDI_ADDRESS_INFO pAI;
	TDI_STATUS err;
	TDI_REQUEST r;
	WORD port;
	DWORD addr;
	uint len;
	int res = 0;

	if (!pNB || !pCBM) {
		res = -SOCKERR_NO_MEMORY;
		goto error;
	}

	if (!(pSock->s_flags & SOCKFL_BOUND)) {
		res = -SOCKERR_NOT_BOUND;
		goto error;
	}

	pAI = (PTDI_ADDRESS_INFO) (pNB + 1);
	pNB->VirtualAddress = (PVOID) pAI;
	pNB->Length = len = sizeof(TDI_ADDRESS_INFO) + 200;

	r.Handle.ConnectionContext = pSock->s_addrobj->addr_hdl;
	r.RequestNotifyObject = &SignalCallback;
	r.RequestContext = (PVOID) pCBM;
	err = pDT->TdiQueryInformationEntry (&r, TDI_QUERY_ADDRESS_INFO, pNB, &len, 0);
	if (err == TDI_PENDING)
		err = WaitCallback(pCBM);
	if (err) {
		res = -err;
		goto error;
	}
	getipport (&pAI->Address, &port, &addr);
	pRegs->Client_EBX = port;
	pRegs->Client_ECX = addr;

error:
	if (pCBM) FreeCallback(pCBM);
	if (pNB) _HeapFree(pNB, 0);
	return res;
}

int
sys_getpeer (HVM hVM, PCRS pRegs, PSOCKET pSock)
{
	if (!(pSock->s_flags & SOCKFL_CONNECTED))
		return -SOCKERR_NOT_CONNECTED;

	pRegs->Client_EBX = pSock->s_remoteport;
	pRegs->Client_ECX = pSock->s_remoteip;
	return 0;
}